@@ -25,19 +25,33 @@ from invoke.vendor.six import text_type, binary_type, PY2
 from invoke.vendor.lexicon import Lexicon
 
 from blessings import Terminal
+from docutils.utils import Reporter
 from enum import Enum
 from invoke import Collection, task, Exit
+import readme_renderer.rst  # transitively required via twine in setup.py
 from releases.util import parse_changelog
 from tabulate import tabulate
+from twine.commands.check import check as twine_check
 
 from .semantic_version_monkey import Version
 
-from ..util import tmpdir
 from ..console import confirm
+from ..environment import in_ci
+from ..util import tmpdir
 
 
 debug = logging.getLogger("invocations.packaging.release").debug
 
+# Monkeypatch readme_renderer.rst so it acts more like Sphinx re: docutils
+# warning levels - otherwise it overlooks (and misrenders) stuff like bad
+# header formats etc!
+# (The defaults in readme_renderer are halt_level=WARNING and
+# report_level=SEVERE)
+# NOTE: this only works because we directly call twine via Python and not via
+# subprocess.
+for key in ("halt_level", "report_level"):
+    readme_renderer.rst.SETTINGS[key] = Reporter.INFO_LEVEL
+
 
 # TODO: this could be a good module to test out a more class-centric method of
 # organizing tasks. E.g.:
@@ -233,6 +247,12 @@ def _converge(c):
     if state.expected_version not in state.tags:
         actions.tag = Tag.NEEDS_CUTTING
 
+    actions.all_okay = (
+        actions.changelog == Changelog.OKAY
+        and actions.version == VersionFile.OKAY
+        and actions.tag == Tag.OKAY
+    )
+
     #
     # Return
     #
@@ -248,8 +268,6 @@ def status(c):
     Doubles as a subroutine, returning the return values from its inner call to
     ``_converge`` (an ``(actions, state)`` two-tuple of Lexicons).
     """
-    # TODO: wants some holistic "you don't actually HAVE any changes to
-    # release" final status - i.e. all steps were at no-op status.
     actions, state = _converge(c)
     table = []
     # NOTE: explicit 'sensible' sort (in rough order of how things are usually
@@ -260,23 +278,43 @@ def status(c):
     return actions, state
 
 
-@task(default=True)
-def all_(c):
+# TODO: thought we had automatic trailing underscore stripping but...no?
+@task(name="all", default=True)
+def all_(c, dry_run=False):
     """
     Catchall version-bump/tag/changelog/PyPI upload task.
+
+    :param bool dry_run:
+        Handed to all subtasks which themselves have a ``dry_run`` flag.
+
+    .. versionchanged:: 2.1
+        Expanded functionality to run ``publish`` and ``push`` as well as
+        ``prepare``.
+    .. versionchanged:: 2.1
+        Added the ``dry_run`` flag.
     """
-    prepare(c)
-    # TODO: below
-    # push(c)
-    # build(c)
-    # publish(c) # TODO: update publish() to accept some of our state and do
-    # things with it like be idempotent?
+    prepare(c, dry_run=dry_run)
+    publish(c, dry_run=dry_run)
+    push(c, dry_run=dry_run)
+    tidelift(c, dry_run=dry_run)
 
 
 @task
-def prepare(c):
+def prepare(c, dry_run=False):
     """
     Edit changelog & version, git commit, and git tag, to set up for release.
+
+    :param bool dry_run:
+        Whether to take any actual actions or just say what might occur. Will
+        also non-fatally exit if not on some form of release branch. Default:
+        ``False``.
+
+    :returns: ``True`` if short-circuited due to all-ok, ``None`` otherwise.
+
+    .. versionchanged:: 2.1
+        Added the ``dry_run`` parameter.
+    .. versionchanged:: 2.1
+        Generate annotated git tags instead of lightweight ones.
     """
     # Print dry-run/status/actions-to-take data & grab programmatic result
     # TODO: maybe expand the enum-based stuff to have values that split up
@@ -284,10 +322,22 @@ def prepare(c):
     # definition too, re: just making them non-enum classes period.
     # TODO: otherwise, we at least want derived eg changelog/version/etc paths
     # transmitted from status() into here...
-    actions, state = status(c)
-    # TODO: unless nothing-to-do in which case just say that & exit 0
-    if not confirm("Take the above actions?"):
-        raise Exit("Aborting.")
+    try:
+        actions, state = status(c)
+    except UndefinedReleaseType:
+        if not dry_run:
+            raise
+        raise Exit(
+            code=0,
+            message="Can't dry-run release tasks, not on a release branch; skipping.",  # noqa
+        )
+    # Short-circuit if nothing to do
+    if actions.all_okay:
+        return True
+    # If work to do and not dry-running, make sure user confirms to move ahead
+    if not dry_run:
+        if not confirm("Take the above actions?"):
+            raise Exit("Aborting.")
 
     # TODO: factor out what it means to edit a file:
     # - $EDITOR or explicit expansion of it in case no shell involved
@@ -295,23 +345,19 @@ def prepare(c):
     # - what else?
 
     # Changelog! (pty for non shite editing, eg vim sure won't like non-pty)
-    if actions.changelog is Changelog.NEEDS_RELEASE:
+    if actions.changelog == Changelog.NEEDS_RELEASE:
         # TODO: identify top of list and inject a ready-made line? Requires vim
         # assumption...GREAT opportunity for class/method based tasks!
         cmd = "$EDITOR {.packaging.changelog_file}".format(c)
-        c.run(cmd, pty=True, hide=False)
-    # TODO: add a step for checking reqs.txt / setup.py vs virtualenv contents
+        c.run(cmd, pty=True, hide=False, dry=dry_run)
     # Version file!
     if actions.version == VersionFile.NEEDS_BUMP:
-        # TODO: suggest the bump and/or overwrite the entire file? Assumes a
-        # specific file format. Could be bad for users which expose __version__
-        # but have other contents as well.
         version_file = os.path.join(
             _find_package(c),
             c.packaging.get("version_module", "_version") + ".py",
         )
         cmd = "$EDITOR {}".format(version_file)
-        c.run(cmd, pty=True, hide=False)
+        c.run(cmd, pty=True, hide=False, dry=dry_run)
     if actions.tag == Tag.NEEDS_CUTTING:
         # Commit, if necessary, so the tag includes everything.
         # NOTE: this strips out untracked files. effort.
@@ -320,11 +366,23 @@ def prepare(c):
             c.run(
                 'git commit -am "Cut {}"'.format(state.expected_version),
                 hide=False,
+                dry=dry_run,
+                echo=True,
             )
         # Tag!
-        c.run("git tag {}".format(state.expected_version), hide=False)
-        # TODO: print something to clarify/confirm tag was cut, if not just
-        # adding echo=True to above
+        c.run(
+            'git tag -a {} -m ""'.format(state.expected_version),
+            hide=False,
+            dry=dry_run,
+            echo=True,
+        )
+    # If top-of-task status check wasn't all_okay, it means the code between
+    # there and here was expected to alter state. Run another check to make
+    # sure those actions actually succeeded!
+    if not dry_run and not actions.all_okay:
+        actions, state = status(c)
+        if not actions.all_okay:
+            raise Exit("Something went wrong! Please fix.")
 
 
 def _release_line(c):
@@ -505,6 +563,14 @@ def _find_package(c):
 def load_version(c):
     package_name = _find_package(c)
     version_module = c.packaging.get("version_module", "_version")
+    # Evict from sys.modules in case we're running at the end of an in-session
+    # edit (eg within prepare()). Otherwise we'll always only see what was
+    # on-disk at first import.
+    # NOTE: must do both the top level package and the version module! Unclear
+    # why. May be due to the specific import strategy; def try using the
+    # cleaner options available under Python 3 when we drop 2.
+    sys.modules.pop("{}.{}".format(package_name, version_module), None)
+    sys.modules.pop(package_name, None)
     # NOTE: have to explicitly give it a bytestr (Python 2) or unicode (Python
     # 3) because https://bugs.python.org/issue21720 HOORAY
     cast = binary_type if PY2 else text_type
@@ -608,11 +674,6 @@ def find_gpg(c):
             return candidate
 
 
-# TODO: open some PRs for twine to push things like dual wheels, better
-# dry-run/cleanroom directory concerns, etc into it.
-# TODO: consider making this idempotent re: checking if the 'current release'
-# already exists on PyPI. Or just hope PyPI response on error is sufficiently
-# useful and trap/print that.
 @task
 def publish(
     c,
@@ -627,7 +688,11 @@ def publish(
     check_desc=False,
 ):
     """
-    Publish code to PyPI or index of choice.
+    Publish code to PyPI or index of choice. Wraps ``build`` and ``publish``.
+
+    This uses the ``twine`` command under the hood, both its pre-upload
+    ``check`` subcommand (which verifies the archives to be uploaded, including
+    checking your PyPI readme) and the ``upload`` one.
 
     All parameters save ``dry_run`` and ``directory`` honor config settings of
     the same name, under the ``packaging`` tree. E.g. say
@@ -635,10 +700,11 @@ def publish(
     archives by default.
 
     :param bool sdist:
-        Whether to upload sdists/tgzs.
+        Whether to upload sdists/tgzs. Default: ``True``.
 
     :param bool wheel:
         Whether to upload wheels (requires the ``wheel`` package from PyPI).
+        Default: ``True``.
 
     :param str index:
         Custom upload index/repository name. See ``upload`` help for details.
@@ -647,11 +713,14 @@ def publish(
         Whether to sign the built archive(s) via GPG.
 
     :param bool dry_run:
-        Skip actual publication step if ``True``.
+        Skip upload step if ``True``.
 
         This also prevents cleanup of the temporary build/dist directories, so
         you can examine the build artifacts.
 
+        Note that this does not skip the ``twine check`` step, just the final
+        upload.
+
     :param str directory:
         Base directory within which will live the ``dist/`` and ``build/``
         directories.
@@ -687,9 +756,15 @@ def publish(
     """
     # Don't hide by default, this step likes to be verbose most of the time.
     c.config.run.hide = False
+    # Including echoing!
+    c.config.run.echo = True
     # Config hooks
+    # TODO: this pattern is too widespread. Really needs something in probably
+    # Executor that automatically does this on our behalf for any kwargs we
+    # indicate should be configurable
     config = c.config.get("packaging", {})
-    index = config.get("index", index)
+    if index is None and "index" in config:
+        index = config["index"]
     if sign is False and "sign" in config:
         sign = config["sign"]
     if dual_wheels is False and "dual_wheels" in config:
@@ -697,6 +772,7 @@ def publish(
     if check_desc is False and "check_desc" in config:
         check_desc = config["check_desc"]
     # Initial sanity check, if needed. Will die usefully.
+    # TODO: remove next backwards incompat release, twine check replaces it
     if check_desc:
         c.run("python setup.py check -r -s")
     # Build, into controlled temp dir (avoids attempting to re-upload old
@@ -705,19 +781,74 @@ def publish(
         # Build default archives
         build(c, sdist=sdist, wheel=wheel, directory=tmp)
         # Build opposing interpreter archive, if necessary
+        # TODO: delete dual wheels when dropping Py2 support
         if dual_wheels:
             if not alt_python:
                 alt_python = "python2"
                 if sys.version_info[0] == 2:
                     alt_python = "python3"
             build(c, sdist=False, wheel=True, directory=tmp, python=alt_python)
-        # Do the thing!
+        # Use twine's check command on built artifacts (at present this just
+        # validates long_description)
+        print(c.config.run.echo_format.format(command="twine check"))
+        failure = twine_check(dists=[os.path.join(tmp, "dist", "*")])
+        if failure:
+            raise Exit(1)
+        # Test installation of built artifacts into virtualenvs (even during
+        # dry run)
+        test_install(c, directory=tmp)
+        # Do the thing! (Maybe.)
         upload(c, directory=tmp, index=index, sign=sign, dry_run=dry_run)
 
 
+@task
+def test_install(c, directory):
+    """
+    Test installation of previously built artifacts found in ``directory``.
+
+    Uses the `venv` module to build temporary virtualenvs.
+    """
+    # TODO: streamline all this in 3.0 when we drop all Py2 support both here
+    # and in downstream repos
+    if PY2:
+        print("WARNING: skipping installation test due to no venv on Python 2")
+        return
+    import venv
+
+    builder = venv.EnvBuilder(with_pip=True)
+    for archive in get_archives(directory):
+        # Skip Python 2 wheels that aren't universal (we're dropping that
+        # entirely soon)
+        if "py2" in archive and "py3" not in archive:
+            continue
+        with tmpdir() as tmp:
+            builder.create(tmp)
+            # Does it install cleanly?
+            # TODO: might be nice to have a further 'can you import whatever it
+            # was' test
+            # TODO: obligatory "is it worth upgrading pip always?"
+            c.run(
+                "{} install --disable-pip-version-check {}".format(
+                    os.path.join(tmp, "bin", "pip"), archive
+                )
+            )
+
+
+def get_archives(directory):
+    # Obtain list of archive filenames, then ensure any wheels come first
+    # so their improved metadata is what PyPI sees initially (otherwise, it
+    # only honors the sdist's lesser data).
+    return list(
+        itertools.chain.from_iterable(
+            glob(os.path.join(directory, "dist", "*.{}".format(extension)))
+            for extension in ("whl", "tar.gz")
+        )
+    )
+
+
 def upload(c, directory, index=None, sign=False, dry_run=False):
     """
-    Upload (potentially also signing) all artifacts in ``directory``.
+    Upload (potentially also signing) all artifacts in ``directory/dist``.
 
     :param str index:
         Custom upload index/repository name.
@@ -729,27 +860,20 @@ def upload(c, directory, index=None, sign=False, dry_run=False):
         Whether to sign the built archive(s) via GPG.
 
     :param bool dry_run:
-        Skip actual publication step if ``True``.
+        Skip actual publication step (and dry-run actions like signing) if
+        ``True``.
 
         This also prevents cleanup of the temporary build/dist directories, so
         you can examine the build artifacts.
     """
-    # Obtain list of archive filenames, then ensure any wheels come first
-    # so their improved metadata is what PyPI sees initially (otherwise, it
-    # only honors the sdist's lesser data).
-    archives = list(
-        itertools.chain.from_iterable(
-            glob(os.path.join(directory, "dist", "*.{}".format(extension)))
-            for extension in ("whl", "tar.gz")
-        )
-    )
+    archives = get_archives(directory)
     # Sign each archive in turn
-    # TODO: twine has a --sign option; but the below is still nice insofar
-    # as it lets us dry-run, generate for web upload when pypi's API is
-    # being cranky, etc. Figure out which is better.
+    # NOTE: twine has a --sign option but it's not quite flexible enough &
+    # doesn't allow you to dry-run or upload manually when API is borked...
     if sign:
         prompt = "Please enter GPG passphrase for signing: "
-        input_ = StringIO(getpass.getpass(prompt) + "\n")
+        passphrase = "" if dry_run else getpass.getpass(prompt)
+        input_ = StringIO(passphrase + "\n")
         gpg_bin = find_gpg(c)
         if not gpg_bin:
             raise Exit(
@@ -760,16 +884,14 @@ def upload(c, directory, index=None, sign=False, dry_run=False):
             cmd = "{} --detach-sign -a --passphrase-fd 0 {{}}".format(
                 gpg_bin
             )  # noqa
-            c.run(cmd.format(archive), in_stream=input_)
+            c.run(cmd.format(archive), in_stream=input_, dry=dry_run)
             input_.seek(0)  # So it can be replayed by subsequent iterations
     # Upload
     parts = ["twine", "upload"]
     if index:
-        index_arg = "--repository {}".format(index)
-    if index:
-        parts.append(index_arg)
+        parts.append("--repository {}".format(index))
     paths = archives[:]
-    if sign:
+    if sign and not dry_run:
         paths.append(os.path.join(directory, "dist", "*.asc"))
     parts.extend(paths)
     cmd = " ".join(parts)
@@ -781,8 +903,46 @@ def upload(c, directory, index=None, sign=False, dry_run=False):
         c.run(cmd)
 
 
-# Stitch together current partway-rewritten stuff into public namespace.
-# TODO: reconsider once fully done; may end up looking a lot like this anyways.
-ns = Collection("release", all_, status, prepare, build, publish)
+@task
+def push(c, dry_run=False):
+    """
+    Push current branch and tags to default Git remote.
+    """
+    # Push tags, not just branches; and at this stage pre-push hooks will be
+    # more trouble than they're worth.
+    opts = "--follow-tags --no-verify"
+    # Dry run: echo, and either tack on git's own dry-run (if not CI) or
+    # dry-run the run() itself (if CI - which probably can't push to the remote
+    # and might thus error uselessly)
+    kwargs = dict()
+    if dry_run:
+        kwargs["echo"] = True
+        if in_ci():
+            kwargs["dry"] = True
+        else:
+            opts += " --dry-run"
+    c.run("git push {}".format(opts), **kwargs)
+
+
+@task
+def tidelift(c, dry_run=False):
+    """
+    Add current latest version to Tidelift & set changelog link.
+    """
+    pass
+
+
+# TODO: still need time to solve the 'just myself pls' problem
+ns = Collection(
+    "release",
+    all_,
+    status,
+    prepare,
+    build,
+    publish,
+    push,
+    test_install,
+    tidelift,
+)
 # Hide stdout by default, preferring to explicitly enable it when necessary.
 ns.configure({"run": {"hide": "stdout"}})
