@@ -8,7 +8,9 @@ import sys
 from invoke.vendor.six import PY2
 from invoke.vendor.lexicon import Lexicon
 from invoke import MockContext, Result, Config, Exit
+from docutils.utils import Reporter
 from mock import Mock, patch
+import pytest
 from pytest import skip
 from pytest_relaxed import trap, raises
 
@@ -24,9 +26,14 @@ from invocations.packaging.release import (
     _release_and_issues,
     _release_line,
     all_,
+    prepare,
+    push,
     build,
     load_version,
+    publish,
     status,
+    upload,
+    ns as release_ns,
 )
 
 
@@ -172,6 +179,14 @@ class load_version_:
     def can_configure_which_module_holds_version_data(self):
         self._expect_version("1.0.1", config_val="otherversion")
 
+    @patch("invocations.packaging.release.sys.modules", wraps=sys.modules)
+    def reloads_version_in_case_edited_during_run(self, modules):
+        # Necessary as default mocks don't mock dunder-attrs
+        modules.__getitem__.return_value._version = Mock(__version__="1.0.0")
+        self._expect_version("1.0.0")
+        modules.pop.assert_any_call("fakepackage._version", None)
+        modules.pop.assert_any_call("fakepackage", None)
+
     def errors_usefully_if_version_module_not_found(self):
         skip()
 
@@ -233,12 +248,6 @@ def _mock_context(self):
     - Where not possible (eg things which must be Python-level and not
       shell-level, such as version imports), mock with the 'mock' lib as usual.
 
-    The MockContext's `run` method has been further mocked by wrapping it in a
-    pass-through `mock.Mock`. It will act like regular `MockContext.run`
-    (returning the result value it's been configured to return) but will be a
-    `mock.Mock` object and thus exhibit all the usual call-tracking attributes
-    and methods such as ``.called``, ``.call_args_list``, etc.
-
     :yields:
         an `invoke.context.MockContext` created & modified as described above.
     """
@@ -367,6 +376,7 @@ Tag +{tag}
                 changelog=Changelog.NEEDS_RELEASE,
                 version=VersionFile.NEEDS_BUMP,
                 tag=Tag.NEEDS_CUTTING,
+                all_okay=False,
             )
             found_actions, found_state = _mock_status(self)
             assert found_actions == actions
@@ -580,18 +590,17 @@ _confirm_false = _confirm(False)
 
 
 # This is shit but I'm too tired and angry right now to give a fuck.
-def _run_all(c, mute=True):
+def _run_prepare(c, mute=True, **kwargs):
     try:
-        return all_(c)
+        return prepare(c, **kwargs)
     except Exit:
         if not mute:
             raise
 
 
-class All:
-    "all_"  # mehhh
+class prepare_:
 
-    # NOTE: just testing the base case of 'everything needs updating',
+    # NOTE: mostly testing the base case of 'everything needs updating',
     # all the permutations are tested elsewhere.
     _branch = "1.1"
     _changelog = "unreleased_1.1_bugs"
@@ -601,7 +610,7 @@ class All:
     @_confirm_false
     def displays_status_output(self, _):
         with _mock_context(self) as c:
-            _run_all(c)
+            _run_prepare(c)
         output = sys.stdout.getvalue()
         for action in (
             Changelog.NEEDS_RELEASE,
@@ -611,17 +620,26 @@ class All:
             err = "Didn't see '{}' text in status output!".format(action.name)
             assert action.value in output, err
 
+    @patch("invocations.packaging.release.status")
+    def short_circuits_when_no_work_to_do(self, status):
+        status.return_value = Lexicon(all_okay=True), Lexicon()
+        with _mock_context(self) as c:
+            # True retval, one call to status(), and no barfing on lack of
+            # run() mocking, all point to the short circuit happening
+            assert _run_prepare(c) is True
+            assert status.call_count == 1
+
     @trap
     @patch("invocations.console.input", return_value="no")
     def prompts_before_taking_action(self, mock_input):
         with _mock_context(self) as c:
-            _run_all(c)
+            _run_prepare(c)
         assert mock_input.call_args[0][0] == "Take the above actions? [Y/n] "
 
     @_confirm_false
     def if_prompt_response_negative_no_action_taken(self, _):
         with _mock_context(self) as c:
-            _run_all(c)
+            _run_prepare(c)
         # TODO: move all action-y code into subroutines, then mock them and
         # assert they were never called?
         # Expect that only the status-y run() calls were made.
@@ -633,38 +651,37 @@ class All:
     @_confirm_true
     def opens_EDITOR_with_changelog_when_it_needs_update(self, _):
         with _mock_context(self) as c:
-            _run_all(c)
+            _run_prepare(c)
             # Grab changelog path from the context config, why not
             path = c.config.packaging.changelog_file
             # TODO: real code should probs expand EDITOR explicitly so it can
             # run w/o a shell wrap / require a full env?
             cmd = "$EDITOR {}".format(path)
-            c.run.assert_any_call(cmd, pty=True, hide=False)
+            c.run.assert_any_call(cmd, pty=True, hide=False, dry=False)
 
     @_confirm_true
     def opens_EDITOR_with_version_file_when_it_needs_update(self, _):
         with _mock_context(self) as c:
-            _run_all(c)
+            _run_prepare(c)
             path = "{}/_version.py".format(FAKE_PACKAGE)
             # TODO: real code should probs expand EDITOR explicitly so it can
             # run w/o a shell wrap / require a full env?
             cmd = "$EDITOR {}".format(path)
-            c.run.assert_any_call(cmd, pty=True, hide=False)
+            c.run.assert_any_call(cmd, pty=True, hide=False, dry=False)
 
     @_confirm_true
     def commits_and_adds_git_tag_when_needs_cutting(self, _):
         with _mock_context(self) as c:
-            _run_all(c)
+            _run_prepare(c)
             version = "1.1.2"  # as changelog has issues & prev was 1.1.1
             # Ensure the commit necessity test happened. (Default mock_context
             # sets it up to result in a commit being necessary.)
             check = 'git status --porcelain | egrep -v "^\\?"'
             c.run.assert_any_call(check, hide=True, warn=True)
             commit = 'git commit -am "Cut {}"'.format(version)
-            # TODO: annotated, signed, etc?
-            tag = "git tag {}".format(version)
+            tag = 'git tag -a {} -m ""'.format(version)
             for cmd in (commit, tag):
-                c.run.assert_any_call(cmd, hide=False)
+                c.run.assert_any_call(cmd, hide=False, dry=False, echo=True)
 
     @_confirm_true
     def does_not_commit_if_no_commit_necessary(self, _):
@@ -672,16 +689,109 @@ class All:
             # Set up for a no-commit-necessary result to check command
             check = 'git status --porcelain | egrep -v "^\\?"'
             c.set_result_for("run", check, Result("", exited=1))
-            _run_all(c)
+            _run_prepare(c)
             # Expect NO git commit
             commands = [x[0][0] for x in c.run.call_args_list]
             assert not any(x.startswith("git commit") for x in commands)
             # Expect git tag
-            c.run.assert_any_call("git tag 1.1.2", hide=False)
+            c.run.assert_any_call(
+                'git tag -a 1.1.2 -m ""', hide=False, dry=False, echo=True
+            )
 
-    def reruns_status_at_end_as_sanity_check(self):
-        # I.e. you might have screwed up editing one of the files...
-        skip()
+    class final_status_check:
+        @_confirm_true
+        @patch("invocations.packaging.release.status")
+        def run_twice_when_not_short_circuiting(self, status, _):
+            status.side_effect = [
+                (
+                    Lexicon(
+                        changelog=Changelog.NEEDS_RELEASE,
+                        version=VersionFile.OKAY,
+                        tag=Tag.OKAY,
+                        all_okay=False,
+                    ),
+                    Lexicon(),
+                ),
+                (Lexicon(all_okay=True), Lexicon()),
+            ]
+            with _mock_context(self) as c:
+                # Mute off - want kaboom if Exit raised
+                _run_prepare(c, mute=False)
+                assert status.call_count == 2
+
+        @_confirm_true
+        @patch("invocations.packaging.release.status")
+        def exits_if_still_not_all_okay(self, status, _):
+            status.side_effect = [
+                (
+                    Lexicon(
+                        changelog=Changelog.NEEDS_RELEASE,
+                        version=VersionFile.OKAY,
+                        tag=Tag.OKAY,
+                        all_okay=False,
+                    ),
+                    Lexicon(),
+                ),
+                (Lexicon(all_okay=False), Lexicon()),
+            ]
+            with _mock_context(self) as c:
+                with pytest.raises(Exit, match=r"Something went wrong"):
+                    _run_prepare(c, mute=False)
+                assert status.call_count == 2
+
+    class dry_run_prepare:
+        @patch("invocations.packaging.release.status")
+        def exits_early_like_non_dry_run_on_all_okay(self, status):
+            status.return_value = Lexicon(all_okay=True), Lexicon()
+            with _mock_context(self) as c:
+                assert _run_prepare(c, dry_run=True) is True
+                assert status.call_count == 1
+
+        @patch("invocations.packaging.release.status")
+        def does_not_fail_fast_on_bad_release_type(self, status):
+            status.side_effect = UndefinedReleaseType
+            with _mock_context(self) as c:
+                _run_prepare(c, dry_run=True)
+
+        @patch("invocations.console.input")
+        def does_not_prompt_to_confirm(self, mock_input):
+            with _mock_context(self) as c:
+                _run_prepare(c, dry_run=True)
+            assert not mock_input.called
+
+        def dry_runs_all_prep_commands(self):
+            # Reminder: default state of mocked context is "everything needs
+            # updates"
+            with _mock_context(self) as c:
+                _run_prepare(c, dry_run=True)
+                dry_runs = [
+                    x[1][0] for x in c.run.mock_calls if x[2].get("dry", False)
+                ]
+                for pattern in (
+                    r"\$EDITOR .*\.rst",
+                    r"\$EDITOR .*_version\.py",
+                    r"git commit.*",
+                    r"git tag -a.*",
+                ):
+                    assert any(re.match(pattern, x) for x in dry_runs)
+
+        @patch("invocations.packaging.release.status")
+        def does_not_run_final_status_check(self, status):
+            # Slight cheat: other actions all actually ok even tho all_okay is
+            # false. means no needing to mock the run() calls etc.
+            status.return_value = (
+                Lexicon(
+                    changelog=Changelog.OKAY,
+                    version=VersionFile.OKAY,
+                    tag=Tag.OKAY,
+                    all_okay=False,
+                ),
+                Lexicon(),
+            )
+            with _mock_context(self) as c:
+                _run_prepare(c, dry_run=True)
+                # The end step was skipped
+                assert status.call_count == 1
 
     # Don't want a full re-enactment of status_ test tree, but do want to spot
     # check that actions not needing to be taken, aren't...
@@ -691,7 +801,7 @@ class All:
         @_confirm_true
         def no_changelog_update_needed_means_no_changelog_edit(self, _):
             with _mock_context(self) as c:
-                _run_all(c)
+                _run_prepare(c)
                 # TODO: as with the 'took no actions at all' test above,
                 # proving a negative sucks - eventually make this subroutine
                 # assert based. Meh.
@@ -702,8 +812,6 @@ class All:
                 )
                 assert cmd not in [x[0][0] for x in c.run.call_args_list], err
 
-    # TODO: rest...
-
 
 # NOTE: yea...this kinda pushes the limits of sane TDD...meh
 # NOTE: possible that the actual codes blessings emits differ based on
@@ -741,9 +849,7 @@ class component_state_enums_contain_human_readable_values:
 
 
 @contextmanager
-def _expect_setuppy(
-    flags, python="python", config=None, yield_rmtree=False
-):
+def _expect_setuppy(flags, python="python", config=None, yield_rmtree=False):
     kwargs = dict(run=True)
     if config is not None:
         kwargs["config"] = config
@@ -894,3 +1000,290 @@ class build_:
                 build(c, clean=False)
             rmtree.assert_any_call("dist", ignore_errors=True)
             rmtree.assert_any_call("build", ignore_errors=True)
+
+
+class upload_:
+    def _check_upload(self, c, kwargs=None, flags=None):
+        """
+        Expect/call upload() with common environment and settings/mocks.
+
+        Returns the full command constructed, typically for further
+        examination.
+        """
+
+        def mkpath(x):
+            return path.join("somedir", "dist", x)
+
+        with patch("invocations.packaging.release.glob") as glob:
+            tgz, whl = mkpath("foo.tar.gz"), mkpath("foo.whl")
+            glob.side_effect = lambda x: [tgz if x.endswith("gz") else whl]
+            # Do the thing!
+            upload(c, "somedir", **(kwargs or {}))
+            glob.assert_any_call(mkpath("*.tar.gz"))
+            glob.assert_any_call(mkpath("*.whl"))
+        self.files = "{} {}".format(whl, tgz)
+        cmd = "twine upload"
+        if flags:
+            cmd += " {}".format(flags)
+        cmd += " {}".format(self.files)
+        return cmd
+
+    def twine_uploads_dist_contents_with_wheels_first(self):
+        c = MockContext(run=True)
+        c.run.assert_called_once_with(self._check_upload(c))
+
+    def may_target_alternate_index(self):
+        c = MockContext(run=True)
+        cmd = self._check_upload(
+            c, kwargs=dict(index="lol"), flags="--repository lol"
+        )
+        c.run.assert_called_once_with(cmd)
+
+    @patch("builtins.print")
+    def dry_run_just_prints_and_ls(self, print):
+        c = MockContext(run=True)
+        cmd = self._check_upload(c, kwargs=dict(dry_run=True))
+        print.assert_any_call("Would publish via: {}".format(cmd))
+        c.run.assert_called_once_with("ls -l {}".format(self.files))
+
+    def allows_signing_via_gpg(self):
+        # Kind of a pain to test :(
+        skip()
+
+
+class Kaboom(Exception):
+    pass
+
+
+class publish_:
+    class base_case:
+        def does_all_the_things(self, fakepub):
+            c, mocks = fakepub
+            # Execution
+            publish(c)
+            # Unhides stdout
+            assert c.config.run.hide is False
+            # Build
+            mocks.build.assert_called_once_with(
+                c, sdist=True, wheel=True, directory="tmpdir"
+            )
+            # Twine check
+            splat = path.join("tmpdir", "dist", "*")
+            mocks.twine_check.assert_called_once_with(dists=[splat])
+            # Install test
+            mocks.test_install.assert_called_once_with(c, directory="tmpdir")
+            # Upload
+            mocks.upload.assert_called_once_with(
+                c, directory="tmpdir", index=None, sign=False, dry_run=False
+            )
+            # Tmpdir cleaned up
+            mocks.rmtree.assert_called_once_with("tmpdir")
+
+        def cleans_up_on_error(self, fakepub):
+            c, mocks = fakepub
+            mocks.build.side_effect = Kaboom
+            with pytest.raises(Kaboom):
+                publish(MockContext(run=True))
+            mocks.rmtree.assert_called_once_with(mocks.mkdtemp.return_value)
+
+        def monkeypatches_readme_renderer(self, fakepub):
+            # Happens at module load time but is just a data structure change
+            import readme_renderer.rst
+
+            assert (
+                readme_renderer.rst.SETTINGS["halt_level"]
+                == Reporter.INFO_LEVEL
+            )
+            assert (
+                readme_renderer.rst.SETTINGS["report_level"]
+                == Reporter.INFO_LEVEL
+            )
+
+    class index:
+        def passed_to_upload(self, fakepub):
+            c, mocks = fakepub
+            publish(c, index="dev")
+            assert mocks.upload.call_args[1]["index"] == "dev"
+
+        def honors_config(self, fakepub):
+            c, mocks = fakepub
+            c.config.packaging = dict(index="prod")
+            publish(c)
+            assert mocks.upload.call_args[1]["index"] == "prod"
+
+        def kwarg_beats_config(self, fakepub):
+            c, mocks = fakepub
+            c.config.packaging = dict(index="prod")
+            publish(c, index="dev")
+            assert mocks.upload.call_args[1]["index"] == "dev"
+
+    class sign:
+        def passed_to_upload(self, fakepub):
+            c, mocks = fakepub
+            publish(c, sign=True)
+            assert mocks.upload.call_args[1]["sign"] is True
+
+        def honors_config(self, fakepub):
+            c, mocks = fakepub
+            c.config.packaging = dict(sign=True)
+            publish(c)
+            assert mocks.upload.call_args[1]["sign"] is True
+
+        def kwarg_beats_config(self, fakepub):
+            c, mocks = fakepub
+            c.config.packaging = dict(sign=False)
+            publish(c, sign=True)
+            assert mocks.upload.call_args[1]["sign"] is True
+
+    class sdist:
+        def defaults_True_and_passed_to_build(self, fakepub):
+            c, mocks = fakepub
+            publish(c)
+            assert mocks.build.call_args[1]["sdist"] is True
+
+        def may_be_overridden(self, fakepub):
+            c, mocks = fakepub
+            publish(c, sdist=False)
+            assert mocks.build.call_args[1]["sdist"] is False
+
+    class wheel:
+        def defaults_True_and_passed_to_build(self, fakepub):
+            c, mocks = fakepub
+            publish(c)
+            assert mocks.build.call_args[1]["wheel"] is True
+
+        def may_be_overridden(self, fakepub):
+            c, mocks = fakepub
+            publish(c, wheel=False)
+            assert mocks.build.call_args[1]["wheel"] is False
+
+    def directory_affects_tmpdir(self, fakepub):
+        c, mocks = fakepub
+        publish(c, directory="explicit")
+        assert not mocks.mkdtemp.called
+        assert mocks.build.call_args[1]["directory"] == "explicit"
+
+    class dry_run:
+        def causes_tmpdir_cleanup_to_be_skipped(self, fakepub):
+            c, mocks = fakepub
+            publish(c, dry_run=True)
+            assert not mocks.rmtree.called
+
+        def causes_tmpdir_cleanup_to_be_skipped_on_exception(self, fakepub):
+            c, mocks = fakepub
+            mocks.build.side_effect = Kaboom
+            with pytest.raises(Kaboom):
+                publish(c, dry_run=True)
+            assert not mocks.rmtree.called
+
+        def passed_to_upload(self, fakepub):
+            c, mocks = fakepub
+            publish(c, dry_run=True)
+            assert mocks.upload.call_args[1]["dry_run"] is True
+
+
+class test_install_:
+    def pip_installs_into_venv(self):
+        skip()
+
+    def aborts_on_python2(self):
+        skip()
+
+    def skips_non_universal_python2_wheels(self):
+        skip()
+
+
+class push_:
+    def pushes_with_follow_tags(self):
+        "git-pushes with --follow-tags"
+        c = MockContext(run=True)
+        push(c)
+        c.run.assert_called_once_with("git push --follow-tags --no-verify")
+
+    @trap
+    def honors_dry_run(self):
+        c = MockContext(run=True)
+        push(c, dry_run=True)
+        c.run.assert_called_once_with(
+            "git push --follow-tags --no-verify --dry-run", echo=True
+        )
+
+    @trap
+    @patch("invocations.environment.os.environ", dict(CIRCLECI="true"))
+    def dry_run_dry_runs_the_invocation_itself_if_in_ci(self):
+        c = MockContext(run=True)
+        push(c, dry_run=True)
+        c.run.assert_called_once_with(
+            "git push --follow-tags --no-verify", echo=True, dry=True
+        )
+
+    @trap
+    @patch("invocations.environment.os.environ", dict(CIRCLECI="true"))
+    def ci_check_only_applies_to_dry_run_behavior(self):
+        # Yes, technically already covered by base tests, but...
+        c = MockContext(run=True)
+        push(c, dry_run=False)
+        c.run.assert_called_once_with("git push --follow-tags --no-verify")
+
+
+class tidelift_:
+    def adds_new_version_with_changelog_link(self):
+        # new version created
+        # release line etc stuff? prob just defaults?
+        # changelog link
+        skip()
+
+    def dry_run_does_not_hit_api(self):
+        skip()
+
+
+class all_task:
+    @patch("invocations.packaging.release.prepare")
+    @patch("invocations.packaging.release.publish")
+    @patch("invocations.packaging.release.push")
+    @patch("invocations.packaging.release.tidelift")
+    def runs_primary_workflow(self, tidelift, push, publish, prepare):
+        c = MockContext(run=True)
+        all_(c)
+        # TODO: this doesn't actually prove order of operations. not seeing an
+        # unhairy way to do that, but not really that worried either...:P
+        prepare.assert_called_once_with(c, dry_run=False)
+        publish.assert_called_once_with(c, dry_run=False)
+        push.assert_called_once_with(c, dry_run=False)
+        tidelift.assert_called_once_with(c, dry_run=False)
+
+    @patch("invocations.packaging.release.prepare")
+    @patch("invocations.packaging.release.publish")
+    @patch("invocations.packaging.release.push")
+    @patch("invocations.packaging.release.tidelift")
+    def passes_through_dry_run_flag(self, tidelift, push, publish, prepare):
+        c = MockContext(run=True)
+        all_(c, dry_run=True)
+        prepare.assert_called_once_with(c, dry_run=True)
+        publish.assert_called_once_with(c, dry_run=True)
+        push.assert_called_once_with(c, dry_run=True)
+        tidelift.assert_called_once_with(c, dry_run=True)
+
+    def bound_to_name_without_underscore(self):
+        assert all_.name == "all"
+
+
+class namespace:
+    def contains_all_tasks(self):
+        names = """
+           all
+           build
+           prepare
+           publish
+           push
+           status
+           test-install
+           tidelift
+        """.split()
+        assert set(release_ns.task_names) == set(names)
+
+    def all_is_default_task(self):
+        assert release_ns.default == "all"
+
+    def hides_stdout_by_default(self):
+        assert release_ns.configuration()["run"]["hide"] == "stdout"
